"""settings parser can be used for parsing input yaml."""

import logging

from requests import Session
from requests.adapters import HTTPAdapter
from web3 import Web3

from Arbie.Actions import ActionTree, RedisState, Store
from Arbie.Contracts import (
    BalancerFactory,
    ContractFactory,
    IERC20Token,
    UniswapFactory,
)
from Arbie.Contracts.contract import Network


def to_network(string: str) -> Network:
    if string.lower() == "mainnet":
        return Network.mainnet
    if string.lower() == "kovan":
        return Network.kovan
    return Network.ropsten


class Keys(object):
    actions = "actions"
    address = "address"
    network = "network"
    store = "store"
    type_key = "type"
    variables = "variables"
    value = "value"
    version = "version"
    web3 = "web3"
    action_tree = "action_tree"
    event = "event"
    redis = "redis"


class VariableParser(object):
    """VariableParser parses settings config and adds variable to store."""

    def __init__(self, config, w3_config):
        self.config = config
        self.w3 = self.set_up_web3(w3_config)

    def add_variables(self, store):
        for name, variable_config in self.config.items():
            store.add(name, self._create_variable(variable_config))
        store.add(Keys.web3, self.w3)

    def set_up_web3(self, config):
        address = config[Keys.address]

        adapter = HTTPAdapter(
            pool_connections=20, pool_maxsize=20, max_retries=10  # noqa: WPS432
        )  # noqa: WPS432
        session = Session()
        session.mount("http://", adapter)
        session.mount("https://", adapter)

        w3 = Web3(Web3.HTTPProvider(address, session=session))

        if not w3.isConnected():
            raise ConnectionError("Web3 is not connected")

        logging.getLogger().info(f"Connected to Node {address}")
        logging.getLogger().info(f"Current block is {w3.eth.blockNumber}")
        return w3

    def set_up_uniswap(self, config):
        return self._set_up_contracts(config, UniswapFactory)

    def set_up_balancer(self, config):
        return self._set_up_contracts(config, BalancerFactory)

    def set_up_token(self, config):
        return self._set_up_contracts(config, IERC20Token).create_token(1)

    def _create_variable(self, variable_config):  # noqa: WPS321
        variable_type = variable_config[Keys.type_key]
        if variable_type == "UniswapFactory":
            return self.set_up_uniswap(variable_config)
        if variable_type == "BalancerFactory":
            return self.set_up_balancer(variable_config)
        if variable_type == "Token":
            return self.set_up_token(variable_config)
        if variable_type == "float":
            return float(variable_config[Keys.value])
        if variable_type == "int":
            return int(variable_config[Keys.value])
        if variable_type == "str":
            return str(variable_config[Keys.value])
        raise TypeError(f"No rule for creating variable if type {variable_type}")

    def _set_up_contracts(self, config, contract):
        factory = ContractFactory(self.w3, contract)
        if Keys.network in config:
            return factory.load_contract(network=to_network(config[Keys.network]))
        return factory.load_contract(address=config[Keys.address])


class SettingsParser(object):
    """Settings Parser is used to configure Arbie."""

    def __init__(self, config):
        self.config = config

    def setup_store(self):
        store = self.store()
        if Keys.variables in self.config:
            self._add_variables(store)
        return store

    def store(self):
        if Keys.store in self.config:
            store_config = self.config[Keys.store]
            redis_state = RedisState(store_config[Keys.address])
            return Store(redis_state)
        return Store()

    def action_tree(self, store):
        if Keys.action_tree in self.config:
            return self._setup_action_tree(store, self.config[Keys.action_tree])

    def _setup_action_tree(self, store, config):
        tree = ActionTree.create(config[Keys.actions], store)
        if Keys.event in config:
            tree.register_event(config[Keys.event])
        return tree

    def _add_variables(self, store):
        variable_parser = VariableParser(
            self.config[Keys.variables], self.config[Keys.web3]
        )
        variable_parser.add_variables(store)



"""Utility functions for interacting with smart contracts."""

import json
from enum import Enum
from typing import Tuple

from pkg_resources import resource_string

from Arbie.async_helpers import run_async


class Network(Enum):
    mainnet = 0
    kovan = 1
    ropsten = 2


def transact(w3, address: str, transaction):
    """Transact a transaction and return transaction receipt."""
    tx_hash = transaction.transact(
        {
            "from": address,
            "gas": 48814000,
        }
    )
    # wait for the transaction to be mined
    return w3.eth.waitForTransactionReceipt(tx_hash, 180)  # noqa: WPS432


class Contract(object):
    """Base class for contracts."""

    def __init__(self, w3, owner_address: str, contract):
        self.w3 = w3
        self.owner_address = owner_address
        self.contract = contract

    def get_address(self) -> str:
        return self.contract.address

    def _transact(self, transaction):
        return transact(self.w3, self.owner_address, transaction)

    def _transact_status(self, transaction) -> bool:
        return self._transact(transaction).status

    def _transact_status_and_contract(self, transaction) -> Tuple[bool, str]:
        tx_receipt = self._transact(transaction)
        return tx_receipt.status, tx_receipt.logs[1].address

    async def _call_async(self, function):
        return await run_async(function.call)


class ContractFactory(object):
    def __init__(self, w3, factory_class: Contract):
        self.w3 = w3

        if (
            factory_class.name is None
            or factory_class.abi is None
            or factory_class.protocol is None
        ):
            raise ValueError(f"{factory_class} dose not contain default parameters")
        self.factory_class = factory_class

    def load_contract(self, owner_address: str = None, **kwargs) -> Contract:
        """Load contract require address or network to be passed in kwargs."""
        address = self._read_address(**kwargs)
        contract = self._load_contract(address)
        return self.factory_class(self.w3, owner_address, contract)

    def deploy_contract(self, owner_address: str, *args) -> Contract:
        contract_address = self._deploy_contract(owner_address, *args)
        contract = self._load_contract(contract_address)
        return self.factory_class(self.w3, owner_address, contract)

    def _deploy_contract(self, deploy_address: str, *args) -> str:
        """Deploy contract and pass on args to contract abi constructor."""
        contract = self.w3.eth.contract(
            abi=self._read_abi(), bytecode=self._read_bytecode()
        )

        transaction = contract.constructor(*args)
        tx_receipt = transact(self.w3, deploy_address, transaction)

        return tx_receipt.contractAddress

    def _read_resource(self, path: str, filename: str) -> str:
        if path is None:
            path = ""
        else:
            path = ".{0}".format(path)

        file_path = "Arbie.resources.contracts{0}".format(path)
        return resource_string(file_path, filename).decode("utf-8")

    def _get_address(self, network: Network):
        json_data = json.loads(self._read_resource(None, "contract_addresses.json"))

        return json_data[self.factory_class.protocol][self.factory_class.abi][
            network.name
        ]

    def _read_abi(self):
        return self._read_resource(
            self.factory_class.protocol, "{0}_abi.json".format(self.factory_class.abi)
        )

    def _read_bytecode(self):
        key = "bytecode"
        filename = "{0}_{1}.json".format(self.factory_class.abi, key)
        json_data = self._read_resource(self.factory_class.protocol, filename)
        return json.loads(json_data)[key]

    def _read_address(self, **kwargs):
        if "address" in kwargs:
            return kwargs.get("address")
        if "network" in kwargs:
            return self._get_address(kwargs.get("network"))

        raise NameError("kwargs does not contain network or address")

    def _load_contract(self, address: str):
        return self.w3.eth.contract(address=address, abi=self._read_abi())



"""Trades contain variables for trading."""

from typing import List

from Arbie.Variables.pool import Pool
from Arbie.Variables.token import Balance, Token


class Trade(object):
    def __init__(self, pool: Pool, token_in: Token, token_out: Token):
        self.pool = pool
        self.token_in = token_in
        self.token_out = token_out


class ArbitrageOpportunity(object):
    def __init__(self, trades: List[Trade], ratio=None):
        self.trades = trades
        self.amount_in = None
        self.profit = None
        self.ratio = ratio

    def __iter__(self):
        return iter(self.trades)

    def __len__(self):
        return len(self.trades)

    def __getitem__(self, i):
        return self.trades[i]

    def set_balance(self, balance: Balance):
        self.balance = balance



"""Main application."""

import logging

from Arbie.settings_parser import SettingsParser


class App(object):
    """App is used for configuring and running Arbie."""

    def __init__(self, config):
        sp = SettingsParser(config)
        self.store = sp.setup_store()
        self.action_tree = sp.action_tree(self.store)

    async def run(self):
        if self.action_tree is None:
            logging.getLogger().warning("No actions given in configuration")
            return
        await self.action_tree.run()

    def stop(self):
        self.action_tree.stop()
